#include "colladaLoader.h"


#include <list>
#include <vector>
#include <string>

#include <iostream>
using namespace std;



ColladaLoader::ColladaLoader()
{
	myVertArray = 0;		
	myNormalArray = 0;	
	myUv0Array = 0;	
	myIndexArray = 0;

	myVertCnt = 0;
	myIndexCnt = 0;
	myPolyCnt = 0;

	myPolyType = Undefined;
}


ColladaLoader::~ColladaLoader()
{
	ReleaseModel();
}

void ColladaLoader::ReleaseModel()
{
	if(myVertArray)
		delete []myVertArray;		
	if(myNormalArray)
		delete []myNormalArray;		
	if(myUv0Array)
		delete []myUv0Array;		
	if(myIndexArray)
		delete []myIndexArray;		

	myVertArray = 0;
	myNormalArray = 0;	
	myUv0Array = 0;	
	myIndexArray = 0;

	myVertCnt = 0;
	myIndexCnt = 0;
	myPolyCnt = 0;

	myPolyType = Undefined;
}

bool ColladaLoader::LoadModel(char* filename)
{
	if(myVertArray)
		ReleaseModel();


	DAE dae;
	domCOLLADA* root = dae.open(filename); // load COLLADA file
	if (!root) {
		cout << "Document import failed.\n";
		return 0;
	}

	daeElement* geomLibNode = root->getDescendant("library_geometries"); // get COLLADA node for geometries

	if (!geomLibNode)
		cout << "No geometries nodes found\n";
	else
	{
		domLibrary_geometries* libGeom=(domLibrary_geometries*)geomLibNode; // typecast to domLibrary_geometries
		domGeometry_Array geomArr=libGeom->getGeometry_array();
		for (size_t i = 0; i < geomArr.getCount(); i++)
		{
			domGeometry* geom=( domGeometry *) geomArr[i];

			domMesh* mesh=geom->getMesh();
			if (mesh) // we are only interested in geometry, which has a mesh
			{
 				domPolylist_Array polyList_array=mesh->getPolylist_array(); // every mash can have multiple groups of polylists/triangles/...
				cout << "polyListArray cnt (polygon groups) = " << polyList_array.getCount() << endl; 
		
				if(polyList_array.getCount()>1)
				{
					cout <<"Supported only one group of polylist\n";
				}

				size_t k=0;
//				for (k=0;k<polyList_array.getCount();k++) // for each group of "polygons"
				{
					domPolylist* polyList=polyList_array[k];

					myPolyCnt = polyList->getCount();
					cout << "polyCnt = " << polyList->getCount() << " (number of polygons in "<<k<<"-th group)" << endl; // number of polygons in k-th group
//					cout << "material= " << polyList->getMaterial() << endl; // material for this group of polygons
					
		 			domPolylist::domVcount* Vcnt = polyList->getVcount(); // list of vertice counts for each polygon
					domListOfUInts vertperpoly=Vcnt->getValue();

					domP* P=polyList->getP();		// list of ID's of units of polygon vertices
					domListOfUInts uintlist=P->getValue();
				
					domInputLocalOffset_Array inputArray=polyList->getInput_array(); // all inputs for geometric primitive set (polylist)
					int inputCnt=inputArray.getCount();

					//pocet prisluchajucich hodnot pre vrchol je zavysli na inputCnt
					//zatial podpora len pre posiciu, normalu, a tex coord0
					if(inputCnt>3)
					{
						cout << "too much atributes for vertex, only 3 attributes allowed\n";
						inputCnt = 3;
					}

					DataBlock* geomPrimitiveInputData[3]; // place for 3 datablocks (POSITION, NORMAL, TEXCOORD_0)

					size_t j;
					for (j=0; j<inputCnt;j++)
					{
						domInputLocalOffset* ilo=inputArray[j];
						int ofs=ilo->getOffset();
						domURIFragmentType source=ilo->getSource();

						daeElementRef sourceRef=source.getElement();

						if ( sourceRef->typeID()==domVertices::ID()) // is this input of type "vertices" ?
						{
							domVertices* vertices=(domVertices*)sourceRef.cast(); // typecast to domVertices

							domInputLocal_Array inLocalArray=vertices->getInput_array();
							for (size_t k=0; k<inLocalArray.getCount();k++)
							{
								domInputLocal* inputLocal=inLocalArray[k];
							
								domURIFragmentType source=inputLocal->getSource(); // source of this input

								daeElementRef sourceRef=source.getElement();

								geomPrimitiveInputData[j]=DataBlock_create(sourceRef,inputLocal->getSemantic(),0);
							}
						}
						else
						{
								if (sourceRef->typeID()==domSource::ID()) // non-vertices
								{
									domSource* src= (domSource*)sourceRef.cast();
									geomPrimitiveInputData[j]=DataBlock_create(sourceRef,ilo->getSemantic(),ilo->getSet());
								}	
						}
					}
				// assembly primitives for this group of polygons

					size_t vertPos_InputID = 0;
					size_t vertNormal_InputID = 0;
					size_t vertUV0_InputID = 0;
					DataBlock* db_vertPos = 0;
					DataBlock* db_vertNormal = 0;
					DataBlock* db_vertUV0 = 0;


					vertPos_InputID=DataBlock_get(geomPrimitiveInputData,3,"POSITION",0); // we are interested in vertex POSITION data
					db_vertPos=geomPrimitiveInputData[vertPos_InputID];

					if(inputCnt>1)
					{
						vertNormal_InputID=DataBlock_get(geomPrimitiveInputData,3,"NORMAL",0); // we are interested in vertex NORMAL data
						db_vertNormal=geomPrimitiveInputData[vertNormal_InputID];
					}

					if(inputCnt>2)
					{
						vertUV0_InputID=DataBlock_get(geomPrimitiveInputData,3,"TEXCOORD",0); // we are interested in vertex TEXCOORD (UV) data
						db_vertUV0=geomPrimitiveInputData[vertUV0_InputID];
					}

			// for simplicity assume, every polygon has 3 vertices
//					size_t totalVertCnt=vertperpoly.getCount()*3;

					// for simplicity assume, each polygon has same number of vertices
					int verticesPerPolygon = vertperpoly[0];
					size_t totalVertCnt=vertperpoly.getCount()*verticesPerPolygon;

					cout << "vertices per polygon = " << verticesPerPolygon << endl;
					switch(verticesPerPolygon)
					{
						case 3:
							myPolyType = Trinagles;
							break;
						case 4:
							myPolyType = Quads; 
							break;

						default:
							myPolyType = Undefined; 
					}

					size_t posSize=3; // assume position has 3 elements (x,y,z), but could be also db_vertPos->unitStride;
					myVertArray=new float[totalVertCnt*posSize]; // vertex position array 

					size_t normalSize=3; // assume normal has 3 elements (nx,ny,nz), but could be also db_vertNormal->unitStride;						
					myNormalArray=new float[totalVertCnt*normalSize]; // vertex normal (nx,ny,nz) array 

					size_t UV0Size=2; // assume texcorrd 0 has 2 elements (u,v), but could be also db_vertUV0->unitStride;						
					myUv0Array=new float[totalVertCnt*UV0Size]; // vertex uv (u,v) array 

					myIndexArray=new int[totalVertCnt]; // index array


					size_t vcnt=0;
                    size_t pos=0;
                    for (size_t j=0; j<vertperpoly.getCount();j++) // for all polygons
                    {
                        pos+=vcnt;
                        vcnt=vertperpoly[j]; // number of vertices for j-th polygon

             
                        for (size_t k=0; k<vcnt;k++) // for all vertices of this polygon
                        {                            
                            /*
                            for every polygon
                                construct all vertices (vertex consists of POSITION, NORMAL, UV0 , ...)
                            vertex construction:
                            1) create new vertex_ID (as there can be only one index for vertex position + vertex normal + vertex uv in a GPU vertex array) 
                            2) read position from source + store it into vertex array
                            3) read normal from source + store it into normal array
                            4) read UV0 from source + store it into TexCoord array, 
                            5) store vertex_ID into indexarray
                            */

                            int pos_ID=uintlist[(pos+k)*inputCnt+vertPos_InputID];                
                            int normal_ID=uintlist[(pos+k)*inputCnt+vertNormal_InputID]; 
                            int UV0_ID=uintlist[(pos+k)*inputCnt+vertUV0_InputID]; 

						// copy position of the vertex into position vertex array
							size_t m;
							if(db_vertPos)
							{
								for (m=0;m<db_vertPos->unitStride;m++)
								{
									myVertArray[myVertCnt*posSize+m]=db_vertPos->data[pos_ID*db_vertPos->unitStride+m];
								}
							}
						// copy normal of the vertex into normal vertex array
							if(db_vertNormal)
							{
								for (m=0;m<db_vertPos->unitStride;m++)
								{
									myNormalArray[myVertCnt*normalSize+m]=db_vertNormal->data[normal_ID*db_vertNormal->unitStride+m];
								}
							}
						// copy normal of the vertex into normal vertex array
							if(db_vertUV0)
							{
								for (m=0;m<db_vertUV0->unitStride;m++)
								{
									myUv0Array[myVertCnt*UV0Size+m]=db_vertUV0->data[UV0_ID*db_vertUV0->unitStride+m];
								}
							}

							myIndexArray[myIndexCnt]=myVertCnt;
							myVertCnt++;
							myIndexCnt++;
						}
					}
			/*
				now we have posArray,normalArray,uv0Array,indexArray for all polygons of one polygongroup together
				we may create VBO's for this polygongroup
			*/
				}
			}		
			else
			{
				cout << "no mesh defined"<< endl;
			}
		}
	}
}




size_t ColladaLoader::DataBlock_get(DataBlock** db,int cnt,xsNMTOKEN semantic,int set)
{
	for (size_t i=0;i<cnt;i++)
	{
		if (db[i]->set==set && strcmp(db[i]->semantic,semantic)==0)
			return i;
	}
	return -1;
}



ColladaLoader::DataBlock* ColladaLoader::DataBlock_create(daeElementRef& sourceRef,xsNMTOKEN semantic, int set)
{
	DataBlock* db=new DataBlock();

	db->semantic=semantic; // semantic (POSITION/NORMAL/TEXCOORD/...) 
	db->set=set;						// set ID (Texcoord 0, Texcoord 1, ...)

	if (sourceRef->typeID()==domSource::ID())
	{
		domSource* src= (domSource*)sourceRef.cast();
		domSource::domTechnique_common* tech=src->getTechnique_common();
		domAccessor* acc=tech->getAccessor();

		db->unitCnt=acc->getCount(); // units in array
		db->unitStride=acc->getStride(); // "distance" between neighbouring units

		domParam_Array params=acc->getParam_array();
// TODO: parse params for correct data assembly

		xsAnyURI accSrc=acc->getSource();
		daeElement* data=accSrc.getElement();
		
		if (data->typeID()==domFloat_array::ID())
		{
			domFloat_array* floatArr=(domFloat_array*)data;
  
			domListOfFloats floats=floatArr->getValue();
			db->data=floats; // data 
		}							
	}
	return db;
}


void ColladaLoader::PrintInfo()
{
	cout << "LoaderInfo:"<<endl<<endl;

	cout << "Polygons ("<<getPolygonsCnt()<<")\n";
	cout << "Vertices ("<<getVertexCnt()<<")\n";
	cout << "Normals ("<<getVertexCnt()<<")\n";
	cout << "Uv ("<<getVertexCnt()<<")\n";
	cout << "Indices ("<<getIndexCnt()<<")\n";
	for(int i=0; i<getVertexCnt(); i++)
	{
		cout <<"V"<<i<<" {"<<(getVertices())[i*3]<<", "<<(getVertices())[i*3+1]<<", "<<(getVertices())[i*3+2]<<"}"<<endl;
		cout <<"N"<<i<<" {"<<(getNormals())[i*3]<<", "<<(getNormals())[i*3+1]<<", "<<(getNormals())[i*3+2]<<"}"<<endl;
		cout <<"U"<<i<<" {"<<(getUv())[i*2]<<", "<<(getUv())[i*2+1]<<"}"<<endl;
	}

	cout << endl;
	for(int i=0; i<getIndexCnt(); i++)
	{
		cout <<"I"<<i<<" - "<<(getIndices())[i]<<endl;
	}
}
