#ifndef _COLLADALOADER_H_
#define _COLLADALOADER_H_

/*
requires library http://sourceforge.net/projects/collada-dom/
collada specification http://www.khronos.org/files/collada_spec_1_5.pdf
*/

/*
Podporovane su vrcholy, normaly a texturove suradnice pre jednu texturu
Podporovane polygony su trojuholniky a stvoruhloniky
!!! cely model musi byt vytvoreny iba z jedneho typu polygonov
*/

#include <dae.h>
#include <dom/domCOLLADA.h>



class ColladaLoader
{
public:
	typedef enum
    {
		Undefined = 0,
		Trinagles = 3,
		Quads = 4, 
    }PolygonType;

	ColladaLoader();
	~ColladaLoader();

	bool LoadModel(char* filename);
	void ReleaseModel();

	float* getNormals()		{ return myNormalArray; }
	float* getVertices()	{ return myVertArray;    }
	float* getUv()			{ return myUv0Array;    }
	int*   getIndices()		{ return myIndexArray;  }

	int getPolygonsCnt(){ return myPolyCnt;  }
	int getVertexCnt()	{ return myVertCnt;  }
	int getIndexCnt()	{ return myIndexCnt; }

	PolygonType getPolyType()	 { return myPolyType; }
	int         getPolyTypeCnt() { return (int) myPolyType; }

	void PrintInfo();

private:
	struct DataBlock
	{
		xsNMTOKEN semantic;  // type POSITION, NORMAL, TEXCOORD, ...
		int set;						// set number of this block type (e.x. TEXCOORD , set = 0,1,2,..)
		int unitStride; // unit size
		int unitCnt;		// number of units
		domListOfFloats data; // data containing values for unit assembly;
	};

	size_t DataBlock_get(DataBlock** db,int cnt,xsNMTOKEN semantic,int set);
	DataBlock* DataBlock_create(daeElementRef& sourceRef,xsNMTOKEN semantic, int set);

private:
	
	float* myVertArray;		// vertex position array 
	float* myNormalArray;	// vertex normal (nx,ny,nz) array 
	float* myUv0Array;		// vertex uv (u,v) array 
	int* myIndexArray;		// index array

	int myVertCnt;
	int myIndexCnt;
	int myPolyCnt;

	PolygonType myPolyType;
};


#endif
