/*
requires library http://sourceforge.net/projects/collada-dom/

collada specification http://www.khronos.org/files/collada_spec_1_5.pdf
*/

#include "colladaLoader.h"

#include <iostream>
using namespace std;

#include <GL/glew.h>
#include <GL/glut.h>


//---------------------------------------------------------------------------

//global parameters

//glut
GLint gLeftButton=0;
GLint gRightButton=0;
GLint gMiddleButton=0;
GLfloat gMouseX=0.0;
GLfloat gMouseY=0.0;
GLfloat gAngleX=0.0;
GLfloat gAngleY=0.0;
GLfloat gPanX=0.0;
GLfloat gPanY=0.0;
GLfloat gZoom=1.0;
GLuint gWinId = 0;
GLuint gResx = 0;
GLuint gResy = 0;
GLfloat gZFar = 10000.0;

//vbo
GLuint gNormalsVB;
GLuint gPositionsVB;
GLuint gTexCoordsVB;
GLuint gTrianglesIB;

//model
ColladaLoader* gLoader=0;

//texture
GLuint gTex0;
typedef unsigned char byte;


//---------------------------------------------------------------------------
// texture stuff
//---------------------------------------------------------------------------


#pragma pack(push)
#pragma pack(1)

typedef struct
{
    byte  identsize;          // size of ID field that follows 18 byte header (0 usually)
    byte  colourmaptype;      // type of colour map 0=none, 1=has palette
    byte  imagetype;          // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

    short colourmapstart;     // first colour map entry in palette
    short colourmaplength;    // number of colours in palette
    byte  colourmapbits;      // number of bits per palette entry 15,16,24,32

    short xstart;             // image x origin
    short ystart;             // image y origin
    short width;              // image width in pixels
    short height;             // image height in pixels
    byte  bpp;               // image bits per pixel 8,16,24,32
    byte  descriptor;         // image descriptor bits (vh flip bits)

    // pixel data follows header

} TGA_HEADER;

#pragma pack(pop)



GLuint loadTGA(const char* fname)
{
	FILE* f=fopen(fname,"rb");
	if (!f) 
		cout<<"Not possible load texture\n"<<endl;

	TGA_HEADER head;
	fread(&head,1,sizeof(TGA_HEADER),f);
	GLvoid* pixels=0;
	GLuint texID=0;
	if (head.imagetype==2) // RGB not packed
	{
		if (head.bpp==24)
		{
			pixels=(unsigned char*)malloc(head.width*head.height*3);
			fread(pixels,head.height,head.width*(head.bpp>>3),f);
			glGenTextures(1,&texID);
			glBindTexture(GL_TEXTURE_2D,texID);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);	// Linear Filtering
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	// Linear Filtering

			if ((head.descriptor & 0x30) == 0) // dolu hlavou, BGR
				glTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,head.width,head.height,0,GL_BGR,GL_UNSIGNED_BYTE,pixels);

		}
	}
	free(pixels);
	fclose(f);
	return texID;
}


bool loadTextures()
{
	gTex0=loadTGA("lena.tga");
	return true;
}


//---------------------------------------------------------------------------
// vbo stuff
//---------------------------------------------------------------------------


bool createVBO()
{
// target = {GL_ARRAY_BUFFER, GL_ELEMENT_ARRAY_BUFFER, GL_PIXEL_UNPACK_BUFFER, GL_PIXEL_PACK_BUFFER}

    glGenBuffers(1, &gPositionsVB); // positions
	glBindBuffer(GL_ARRAY_BUFFER, gPositionsVB);
	int size = gLoader->getVertexCnt()*3*sizeof(float);
	glBufferData(GL_ARRAY_BUFFER, size, gLoader->getVertices(), GL_STATIC_DRAW_ARB);


    glGenBuffers(1, &gNormalsVB); // normals
	glBindBuffer(GL_ARRAY_BUFFER, gNormalsVB);
	size = gLoader->getVertexCnt()*3*sizeof(float);
	if(gLoader->getNormals())
		glBufferData(GL_ARRAY_BUFFER, size, gLoader->getNormals(), GL_STATIC_DRAW_ARB);


    glGenBuffers(1, &gTexCoordsVB); // texture coordinates
	glBindBuffer(GL_ARRAY_BUFFER, gTexCoordsVB);
	size = gLoader->getVertexCnt()*2*sizeof(float);
	glBufferData(GL_ARRAY_BUFFER, size, gLoader->getUv(), GL_STATIC_DRAW_ARB);


	glGenBuffers(1, &gTrianglesIB); // Index Buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gTrianglesIB);
	size = gLoader->getIndexCnt()*sizeof(int);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, size, gLoader->getIndices(), GL_STATIC_DRAW_ARB);

	return true;
}

bool deleteVBO()
{
	glDeleteBuffers(1,&gPositionsVB);
	glDeleteBuffers(1,&gNormalsVB);
	glDeleteBuffers(1,&gTexCoordsVB);
	glDeleteBuffers(1,&gTrianglesIB);
	return true;
}

//---------------------------------------------------------------------------
// gl stuff
//---------------------------------------------------------------------------


void InitGL()
{
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glClearDepth(1.0);
    glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(45.0,640.0/480.0,0.1, gZFar);


	//flat or smooth shading
	glShadeModel(GL_FLAT );

	//wireframe or filled polygons
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	
	GLfloat LightAmbient[]= { 0.5f, 0.5f, 0.5f, 1.0f }; 
	GLfloat LightDiffuse[]= { 1.0f, 0.4f, 0.8f, 1.0f };
	GLfloat LightSpecular[]= { 0.3, 1.0f, 0.2f, 1.0f };
	GLfloat LightPosition[]= { 0.0f, 0.0f, 0.0f, 1.0f };

	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);	
	glLightfv(GL_LIGHT1, GL_SPECULAR, LightSpecular);	
	glLightfv(GL_LIGHT1, GL_POSITION, LightPosition);
	
	glEnable(GL_LIGHT1);	
	glEnable(GL_LIGHTING);	


    //enable glew library for using gl extensions
	if (GLEW_OK != glewInit())
    {
		fprintf(stderr,"Error glew initialization failed.\n");
    }
}


//---------------------------------------------------------------------------
// glut stuff
//---------------------------------------------------------------------------


void Redraw()
{
//	int time = glutGet(GLUT_ELAPSED_TIME);

	glLoadIdentity();
	glTranslatef(0,0,-2.5);
	glTranslated(gPanX,gPanY,gZoom);
//	glRotatef((float)time/100.0f,0,1,1);
	glRotatef(gAngleX,0,1,0);
	glRotatef(gAngleY,1,0,0);


	glClearColor(0.0,0.0,0.0,1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glColor4f(1.0,1.0,1.0,1.0);

	glBindBufferARB( GL_ARRAY_BUFFER_ARB, gPositionsVB);	// bind ARRAY BUFFER with glBuffer of vertex positions
	glVertexPointer( 3, GL_FLOAT, 0, (char *) NULL );		// Set The Vertex Pointer To bound ARRAY BUFFER

	glBindBufferARB( GL_ARRAY_BUFFER_ARB, gTexCoordsVB );	// bind ARRAY BUFFER with glBuffer of texture coordinates
	glTexCoordPointer( 2, GL_FLOAT, 0, (char *) NULL );		// Set The TexCoord Pointer To bound ARRAY BUFFER

	glBindBufferARB( GL_ARRAY_BUFFER_ARB, gNormalsVB );		// bind ARRAY BUFFER with glBuffer of normals
	glNormalPointer( GL_FLOAT, 0, (char *) NULL );		// Set The Normals Pointer To bound ARRAY BUFFER

	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, gTrianglesIB);	// bind ELEMENT ARRAY BUFFER with glBuffer of indices


	glEnableClientState( GL_VERTEX_ARRAY );					// Enable Vertex Arrays
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );			// Enable Texture Coord Arrays
	glEnableClientState( GL_NORMAL_ARRAY );					// Enable Normal Arrays


// render indexed primitives using indices stored in bound ELEMENT ARRAY BUFFER (last parameter is NULL)
	if(gLoader->getPolyType() == ColladaLoader::Trinagles)
		glDrawRangeElements(GL_TRIANGLES, 0, gLoader->getIndexCnt(), gLoader->getIndexCnt(), GL_UNSIGNED_INT, NULL);
	if(gLoader->getPolyType() == ColladaLoader::Quads)
		glDrawRangeElements(GL_QUADS, 0, gLoader->getIndexCnt(), gLoader->getIndexCnt(), GL_UNSIGNED_INT, NULL);


	glDisableClientState( GL_VERTEX_ARRAY );				// Disable Vertex Arrays
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );			// Disable Texture Coord Arrays
	glDisableClientState( GL_NORMAL_ARRAY );				// Disable Normal Arrays

    glutSwapBuffers();
}



void ReSize(int width_in, int height_in)
{
    glViewport(0, 0, width_in, height_in);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(45.0, (float)width_in/(float)height_in ,0.1, gZFar);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	gResx = width_in;
	gResy = height_in;

	Redraw();
}


void MouseMove(int x_in, int y_in)
{
 if(gLeftButton)
 {
    gAngleX+=x_in-gMouseX;
    gAngleY-=gMouseY-y_in;
 }

 if(gRightButton)
 {
    gPanX-=(x_in-gMouseX)*(gZoom/1000);
    gPanY-=(gMouseY-y_in)*(gZoom/1000);
 }

 if(gMiddleButton)
 {
    gZoom += ((x_in-gMouseX) + (y_in-gMouseY))*(gZoom/500);
 }


 gMouseX=x_in;
 gMouseY=y_in;
}


void MousePress(int button_in, int state_in, int x_in, int y_in)
{
 if(button_in == GLUT_LEFT_BUTTON)
    gLeftButton = 1;
 else
    gLeftButton = 0;

 if(button_in == GLUT_RIGHT_BUTTON)
    gRightButton = 1;
 else
    gRightButton = 0;

 if(button_in == GLUT_MIDDLE_BUTTON)
    gMiddleButton = 1;
 else
    gMiddleButton = 0;


 if(state_in == GLUT_DOWN)
 {
    gMouseX=x_in;
    gMouseY=y_in;
 }
 else
 {
 }
}

void KeyPressed(unsigned char key_in, int x_in, int y_in)
{
    switch(key_in)
	{
		case 'w':
			{
				static bool wire = 1;
				if( wire )
					glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
				else
					glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
				wire = !wire;
			}
			break;

		case 's':
			{
				static bool shading = 1;
				if( shading )
					glShadeModel(GL_SMOOTH);
				else
					glShadeModel(GL_FLAT);
				shading = !shading;
			}
			break;

		case 'l':
			{
				static bool light = 1;
				if( light )
					glDisable(GL_LIGHTING);	
				else
					glEnable(GL_LIGHTING);	
				light = !light;
			}
			break;
		
		case 't':
			{
				static bool text = 1;
				glBindTexture(GL_TEXTURE_2D, gTex0);

				if(text)
					glEnable(GL_TEXTURE_2D);
				else
					glDisable(GL_TEXTURE_2D);
				text = !text;
			}
			break;


		case 27:
			{
                glutDestroyWindow(gWinId);
                deleteVBO();
				delete gLoader;
			    exit(0);
			}
			break;
	}
}



void usage()
{
	cout <<"\nKey shortcats:\n";
	cout <<"\t w - wireframe on/off\n";
	cout <<"\t t - texture on/off\n";
	cout <<"\t l - lightning on/off\n";
	cout <<"\t s - shading on/off\n";
}


int main(int argc, char **argv) 
{
    //set glut window settings
    glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitWindowSize(640, 480);
	glutInitWindowPosition(100, 100);
    //create glut window
    gWinId = glutCreateWindow("Collada loader example");

    //set glut contorling function
	glutDisplayFunc(&Redraw);
	glutIdleFunc(&Redraw);
	glutReshapeFunc(&ReSize);
	glutKeyboardFunc(&KeyPressed);
	glutMouseFunc(&MousePress);
	glutMotionFunc(&MouseMove);


	InitGL();
//	loadTextures();

	gLoader = new ColladaLoader;
//	gLoader->LoadModel("box.dae"); gZoom = -3;
	gLoader->LoadModel("monkey.dae"); gZoom= -10;

	loadTextures();
	createVBO();
	usage();

	//enable main loop
	glutMainLoop();

	delete gLoader;

	return 0;
}
